unit ThreadsLift;

interface

uses
  SysUtils, Classes, Math, IdTCPClient, ComCtrls;

const
  stateUnknown = 0;
  doorOpen = 1;
  doorClosed = 2;

  dirUp = 1;
  dirDown = 2;

type
  TLiftShaft = record
    door: Integer;
    dispatchedTo: Integer;
    speed: Integer;
    meter: Integer;
    direction: Integer;
    dests: set of Byte;
    clearDest: Integer;
  end;

  TLiftThread = class(TThread)
  private
    shafts: array of TLiftShaft;
    levelHeight: Integer;
    levels: Integer;
    maxSpeed: Integer;
    maxAccel: Integer;
    accelLength: Integer;
    waitups, waitdowns: set of Byte;
  protected
    procedure Execute; override;
    function HandleCommand(cmd: string; params: TStrings): integer;
    procedure SendCommand(cmd: string);

    procedure DispatchLift(lift: Integer; level: Integer);
    function GetLevelPos(level: Integer): Integer;
    procedure AdjustLiftSpeed(lift: Integer);
    procedure RequestLift(level: Integer; dir: Integer);
    procedure AssignJob();

    procedure liftShafts(count: Integer);
    procedure liftClosed(index: Integer);
    procedure liftOpened(index: Integer);
    procedure liftLevelButtonUp(level: Integer);
    procedure liftLevelButtonDown(level: Integer);
    procedure liftMeter(index: Integer; meter: Integer);
    procedure liftSpeed(index: Integer; meter: Integer);
    procedure liftLevelHeight(height: Integer);
    procedure liftLevel(lift, level: Integer);
    procedure liftLevels(count: Integer);
    procedure liftMaxSpeed(num: Integer);
    procedure liftMaxAccel(num: Integer);
    procedure liftCabbinButton(lift, button: Integer);
  public
    client: TIdTCPClient;
    log: TListView;
  end;

implementation

{ TLiftThread }

procedure TLiftThread.Execute;
var
  Command: string;
  CmdName: string;
  logitem: TListItem;
  params: TStringList;
  i, j: Integer;
begin
  // The client is supposed to be connected and ready to transmit / receive.
  // Initialize the thread...
  SendCommand('param');

  params := TStringList.Create;
  while not Terminated  and client.Connected do begin
    try
      Command := client.ReadLn();

      // Parse the command name...
      params.Clear;
      i := Pos(' ', Command);
      if (i = 0) then
        CmdName := Command
      else begin
        CmdName := Copy(Command, 1, i - 1);
        i := i + 1;
        j := i;
        while i <= Length(Command) do begin
          if Command[i] = ' ' then begin
            params.Add(Copy(Command, j, i - j));
            j := i + 1;
          end;
          i := i + 1;
        end;
        if i <> j then
          params.Add(Copy(Command, j, i - j));
      end;

      logitem := log.Items.Add;
      logitem.Caption := Command;
      logitem.SubItems.Add('Server');

      try
        case HandleCommand(CmdName, params) of
          2: logitem.SubItems.Add('Not enough parameters');
          1: logitem.SubItems.Add('Handled');
          0: logitem.SubItems.Add('Ignored');
        end;
      except
        logitem.SubItems.Add('Exception thrown');
      end;

    except
    end;
  end;

  params.Free;
  Terminate;
end;

function TLiftThread.HandleCommand(cmd: string; params: TStrings): Integer;
begin
  Result := 0;

  if cmd = 'shafts' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftShafts(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'closed' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftClosed(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'opened' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftOpened(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'levelbuttonup' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftLevelButtonUp(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'levelbuttondn' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftLevelButtonDown(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'levelheight' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftLevelHeight(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'levels' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftLevels(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'maxspeed' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftMaxSpeed(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'maxaccel' then begin
    if (params.Count < 1) then
      Result := 2
    else begin
      liftMaxAccel(StrToInt(params[0]));
      Result := 1;
    end;
  end;

  if cmd = 'speed' then begin
    if (params.Count < 2) then
      Result := 2
    else begin
      liftSpeed(StrToInt(params[0]), StrToInt(params[1]));
      Result := 1;
    end;
  end;

  if cmd = 'meter' then begin
    if (params.Count < 2) then
      Result := 2
    else begin
      liftMeter(StrToInt(params[0]), StrToInt(params[1]));
      Result := 1;
    end;
  end;

  if cmd = 'cabinbutton' then begin
    if (params.Count < 2) then
      Result := 2
    else begin
      liftCabbinButton(StrToInt(params[0]), StrToInt(params[1]));
      Result := 1;
    end;
  end;

  if cmd = 'level' then begin
    if (params.Count < 2) then
      Result := 2
    else begin
      liftLevel(StrToInt(params[0]), StrToInt(params[1]));
      Result := 1;
    end;
  end;
end;

procedure TLiftThread.liftClosed(index: Integer);
begin
  shafts[index].door := doorClosed;
  AdjustLiftSpeed(index);
end;

procedure TLiftThread.liftShafts(count: Integer);
begin
  SetLength(shafts, count);
end;

procedure TLiftThread.liftOpened(index: Integer);
begin
  shafts[index].door := doorOpen;
end;

procedure TLiftThread.SendCommand(cmd: string);
var
  logitem: TListItem;
begin
  client.WriteLn(cmd);
  logitem := log.Items.Add;
  logitem.Caption := cmd;
  logitem.SubItems.Add('Client');
end;

procedure TLiftThread.liftLevelButtonDown(level: Integer);
begin
  SendCommand(Format('indidnon %d', [level]));
  DispatchLift(0, level);
end;

procedure TLiftThread.liftLevelButtonUp(level: Integer);
begin
  SendCommand(Format('indiupon %d', [level]));
end;

procedure TLiftThread.liftMeter(index, meter: Integer);
begin
  shafts[index].meter := meter;
  AdjustLiftSpeed(index);
end;

procedure TLiftThread.liftSpeed(index, meter: Integer);
begin
  shafts[index].speed := meter;
end;

procedure TLiftThread.liftLevelHeight(height: Integer);
begin
  levelHeight := height;
end;

procedure TLiftThread.DispatchLift(lift, level: Integer);
begin
  shafts[lift].dispatchedTo := GetLevelPos(level);
  shafts[lift].clearDest := level;
  AdjustLiftSpeed(lift);
end;

function TLiftThread.GetLevelPos(level: Integer): Integer;
begin
  Result := (levels - level - 1) * levelHeight;
end;

procedure TLiftThread.liftLevels(count: Integer);
var
  i: Integer;
begin
  levels := count;
  for i := 0 to Length(shafts) - 1 do begin
    FillChar(shafts[i], sizeof(shafts[i]), 0);
    shafts[i].meter := -1;
    sendCommand(Format('close %d', [i]));
    sendCommand(Format('speed %d 1', [i]));
    DispatchLift(i, 0);
  end;
end;

procedure TLiftThread.liftMaxAccel(num: Integer);
begin
  maxAccel := num;
  accelLength := maxSpeed * maxSpeed div maxAccel div 2; 
end;

procedure TLiftThread.liftMaxSpeed(num: Integer);
begin
  maxSpeed := num;
end;

procedure TLiftThread.AdjustLiftSpeed(lift: Integer);
var
  accel: Integer;
  newSpeed: Integer;
  accelLength: Integer;
begin
  if (levels = 0) or (shafts[lift].meter = -1) or (shafts[lift].door <> doorClosed) or (shafts[lift].dispatchedTo = -1) then Exit;

  accelLength := shafts[lift].speed * shafts[lift].speed div maxAccel div 2;

  if (abs(shafts[lift].dispatchedTo - shafts[lift].meter) <= accelLength) then begin
    accel := Max(-maxAccel, -shafts[lift].speed);
  end else
    accel := shafts[lift].dispatchedTo - shafts[lift].speed - shafts[lift].meter;

  if accel > maxAccel then
    accel := maxAccel;
  if accel < -maxAccel then
    accel := -maxAccel;

  newSpeed := shafts[lift].speed + accel;
  if newSpeed > maxSpeed then newSpeed := maxSpeed;
  if newSpeed < -maxSpeed then newSpeed := -maxSpeed;

  SendCommand(Format('speed %d %d', [lift, newSpeed]));
  shafts[lift].speed := 0;

  if newSpeed = 0 then begin
    shafts[lift].dispatchedTo := -1;
    if (shafts[lift].clearDest <> -1) then shafts[lift].dests := shafts[lift].dests - [shafts[lift].clearDest];
    if shafts[lift].dests = [] then shafts[lift].direction := stateUnknown;
  end;
end;

procedure TLiftThread.liftLevel(lift, level: Integer);
begin
  SendCommand(Format('shaftinfo %d %d', [lift, level]));
end;

procedure TLiftThread.RequestLift(level, dir: Integer);
var
  levelpos: Integer;
  i: Integer;
begin
  // Find the neared lift with correct direction
  levelpos := GetLevelPos(level);
  for i := 0 to Length(shafts) - 1 do begin
    if (dir = dirUp) and (shafts[i].direction <> dirDown) and (shafts[i].meter > levelpos) then break;
    if (dir = dirDown) and (shafts[i].direction <> dirUp) and (shafts[i].meter < levelpos) then break;
  end;

  if levelpos = Length(shafts) then begin

  end;
end;

procedure TLiftThread.AssignJob;
var
  i, j: Integer;
begin
  for i := 0 to Length(shafts) - 1 do begin
    if (shafts[i].dispatchedTo = -1) then begin
      if (shafts[i].direction = dirUp) then begin
        for j := shafts[i].clearDest + 1 to levels - 1 do
          if (j in shafts[i].dests) then begin
            DispatchLift(i, j);
            break;
          end;
        if j = levels then shafts[i].direction := stateUnknown;
      end;
      if (shafts[i].direction = dirDown) then begin
        for j := 0 to shafts[i].clearDest - 1 do
          if (j in shafts[i].dests) then begin
            DispatchLift(i, j);
            break;
          end;
        if j = shafts[i].clearDest then shafts[i].direction := stateUnknown;
      end;
    end;
  end;
end;

procedure TLiftThread.liftCabbinButton(lift, button: Integer);
begin
  SendCommand(Format('indion %d %d', [lift, button]));
end;

end.
